#!/bin/sh
PACKAGE="AT&T djvu compressor 2.0 [(c) AT&T 1998]"
PLATFORM="Linux"
BIN_PERM=555
MAN_PERM=444

ME=`basename $0`

# Some systems do not have "line"
line() {
	read temp
	echo $temp
}

# echo a line without a terminator
echon() {	# arguments to echo...
	if [ "x`echo -n t`" = "x-n t" ]; then
		echo "${*}\c"
	else
		echo -n "$*"
	fi
}

# get a yes/no response from the user
y_n() {	# default "message"
	export Y_N
	Y_N=
	while [ -z "$Y_N" ]; do
		echon "$2 "
		if [ "x$1" = 'xy' ]; then
			echon "(Y/n) ? "
		else
			echon "(y/N) ? "
		fi
		V=`line`
		if [ -z "$V" ]; then
			V="$1"
		fi
		case "$V" in
		y*|Y*)	Y_N=y;;
		n*|N*)	Y_N=n;;
		*)	echo "Please answer either yes or no";;
		esac
	done
}

getdir() {	# default "message" optional
	export DIR
	if [ -n "$3" ]; then
		y_n n "Should I $2"
		if [ "x$Y_N" = 'xn' ]; then
			DIR=
			return
		fi
	fi
	DIR=$1
	while true; do
		echon "Where do I ${2}? [${DIR}]: "
		V=`line`
		if [ -n "$V" ]; then
			DIR=`echo $V | sed -e s+~+$HOME+`
		fi
		if [ ! -d "$DIR" ]; then
			mkdir -p "$DIR"
		fi
		if [ -w "$DIR" ]; then
			break
		else
			echo "$ME: $DIR is not writable"
			echo
		fi
	done
}

installem() { # fromdir todir perm files...
	fromdir=$1
	shift
	todir="$1"
	if [ -z "$todir" ]; then
		return
	fi
	shift
	perm=$1
	shift
	echon "Installing in ${todir}: "
	for F in $*; do
		# Special processing for ".in" files
		case "$F" in
		*.in)	G=`basename $F .in`
			echon "$G "
			rm -f $todir/$G
			sed -e "s+#LIB#+$LIBDIR+" -e "s+#BIN#+$BINDIR+" $fromdir/$F >$todir/$G
			chmod $perm $todir/$G
			;;
		*.so.[0-9]*.*)
			echon "$F "
			cp $fromdir/$F $todir/$F.new
			chmod $perm $todir/$F.new
			rm -f $todir/$F
			mv $todir/$F.new $todir/$F
			G=`echo $F|sed -e 's+\([.]so[.][0-9][0-9]*\)[.].*+\1+'`
			ln -sf $F $todir/$G
			GG=`echo $G|sed -e 's,[.]so[.][0-9][0-9]*$,.so,'`
			ln -sf $G $todir/$GG
			;;
		*.so.*)
			;;
		*.so)
			;;
		*)
			G=`(cd $fromdir >>/dev/null 2>>/dev/null;ls -la "$F"|sed -n -e "s,^l.* $F -> \(.*\)$,\\1,p")`
			if [ -z "$G" ] ; then
				echon "$F "
				rm -f $todir/$F
				cp $fromdir/$F $todir
				chmod $perm $todir/$F
			else
				echon "$F($G) "
				rm -f $todir/$F
				(cd $todir >>/dev/null 2>>/dev/null;ln -sf $G $F)
			fi
			;;
		esac
	done
	echo
}

#
# Ask the user where to install everything
#
echo
echo "This is $PACKAGE for $PLATFORM"
echo
echo
echo "This software is covered by a license agreement."
echon "Please hit ENTER and read it now."
line
more LICENSE
echo
y_n n "Do you accept the terms of the license agreement"
if [ "x$Y_N" = 'xn' ]; then
	echo
	echo "Please delete this software package NOW"
	echo
	exit 1
fi
if [ -r archive.tar ]
then
  echo "Extracting the archive.  Please wait"
  if ( tar xf archive.tar ) 
  then
    rm -f archive.tar
  else
    echo "Failed to extract archive" 1>&2
    exit 1
  fi
fi
y_n n "Would you like to install this package now?"
if [ "x$Y_N" = 'xn' ]; then
        echo "You man run $0 again later to complete the installation."
	exit 0
fi
BIN_FILES=`(cd ./bin >>/dev/null 2>>/dev/null;ls|sed -e '/Makefile/d' -e '/[.]in$/d')`
CONTRIB_FILES=`(cd ./contrib >>/dev/null 2>>/dev/null;ls|sed -e '/Makefile/d' -e '/[.]in$/d')`
CONTRIB_MAN_FILES=`(cd ./contribman >>/dev/null 2>>/dev/null;ls|sed -e '/Makefile/d' -e '/[.]in$/d')`
SOLIB_FILES=`(cd ./lib >>/dev/null 2>>/dev/null; echo *.so.[0-9]*.* )`
LIB_FILES=`(cd ./lib >>/dev/null 2>>/dev/null;ls|sed -e '/Makefile/d' -e 's,.in$,,' -e '/[^ ]*[.]so/d' -e '/[^ ]*[.]so[.][^ ]*/d')`
MAN_FILES=`(cd ./man >>/dev/null 2>>/dev/null;ls|sed -e '/Makefile/d' -e '/[.]in$/d')`
echo
echo "You will be prompted for the locations to install the"
echo "software.  This install script will make any required"
echo "directories."
echo
echo "The programs include:"
echo "$BIN_FILES"
echo
getdir /usr/bin "install the programs"
BINDIR=$DIR
DEF=`echo $BINDIR | sed -e 's,//*$,,' -e 's,^bin$,,' -e 's,/bin$,,' -e 's,//*$,,' -e 's+$+/man/man1+'`
getdir $DEF "put man pages"
MANDIR=$DIR
DEF=`echo $BINDIR | sed -e 's,//*$,,' -e 's,^bin$,,' -e 's,/bin$,,' -e 's,//*$,,' -e 's+$+/lib+'`
getdir $DEF "install the shared library"
SOLIBDIR=$DIR
export SOLIBDIR
DEF=`echo $BINDIR | sed -e 's,//*$,,' -e 's,^bin$,,' -e 's,/bin$,,' -e 's,//*$,,' -e 's+$+/lib/djvu+'`
getdir $DEF "install the support library commands"
LIBDIR=$DIR
export LIBDIR
echo
echo "The 'contrib' programs are unsupported additions to"
echo "this release of the software"
echo
echo "Currently in 'contrib' are:"
echo "$CONTRIB_FILES"
echo
getdir $BINDIR "install the contrib programs" optional
CONTRIBDIR=$DIR
CONTRIBMANDIR=
if [ -n "$CONTRIBDIR" ]; then
	getdir $MANDIR "install the contrib man pages"
	CONTRIBMANDIR=$DIR
fi

echo

#
# Now, actually do the installing
#

installem bin "$BINDIR" $BIN_PERM $BIN_FILES
installem lib "$SOLIBDIR" $BIN_PERM $SOLIB_FILES
installem lib "$LIBDIR" $BIN_PERM $LIB_FILES
installem man "$MANDIR" $MAN_PERM $MAN_FILES

for i in djvuencode djvudecode djvucmp ; do
  if [ -x "$BINDIR/$i"-shared ] ; then
    if ( LD_LIBRARY_PATH="$SOLIBDIR";export LD_LIBRARY_PATH;"$BINDIR/$i"-shared -h >/dev/null 2>/dev/null ) ; then
      echo "$i"-shared "works, using it as $i"
      rm -f "$BINDIR/$i"
      ( cd "$BINDIR" >>/dev/null 2>>/dev/null;mv "$i"-shared "$i" )
    else
      echo "$i"-shared "doesn't work, using static"
      rm -f "$BINDIR/$i"-shared
    fi
  fi
done

installem contrib "$CONTRIBDIR" $BIN_PERM $CONTRIB_FILES
installem contribman "$CONTRIBMANDIR" $MAN_PERM $CONTRIB_MAN_FILES
echo "----------------------------------------------------------------"
echo " "
if [ -x /sbin/ldconfig ] ; then
  /sbin/ldconfig 2>/dev/null >/dev/null
else
  ldconfig 2>/dev/null >/dev/null
fi
if ( $BINDIR/djvuencode -h >/dev/null 2>/dev/null ) ; then
  echo "Installation complete"
else
  echo "Now add:"
  echo "  LD_LIBRARY_PATH=$SOLIBDIR;export LD_LIBRARY_PATH"
  echo "to your profile, to complete this installation."
fi
exit 0

